local lineextradetails = {}

local _utils = commonapi.utils
local _log = _utils.logPrep("commonapi2.lineextradetails")
local _syncdata = commonapi._syncdata

local linesextradetails_sync =
	_syncdata.SyncObject:new(
	{
		key = "linesextradetails",
		data = {}
	}
)

local datastorage = linesextradetails_sync:getDataOrInit({})


function lineextradetails.applySavegameData(data)
	if (type(data) ~= "table") then
		return
	end
	-- TODO: Better Version check
	if (data.version == 1) then
		linesextradetails_sync:syncData(data)
		datastorage = linesextradetails_sync:getData()
		if (commonapi._native ~= nil and commonapi._native.LabelSystem ~= nil) then
			commonapi._native.LabelSystem.updateData(datastorage)
		end
		-- TODO: Verifiy Data?
	end
end

function lineextradetails.notifyDataChange()
	_log("notifyDataChange")
	datastorage = linesextradetails_sync:getData()
end


function lineextradetails.getStateData()
	return linesextradetails_sync:getData()
end

function lineextradetails.syncData()
	_log("syncData")
	lineextradetails.verifyLines()

	linesextradetails_sync:syncData()
	datastorage = linesextradetails_sync:getData()
	if (commonapi._native.LabelSystem ~= nil) then
		commonapi._native.LabelSystem.updateData(datastorage)
	end
	commonapi_gamescript.saveData()
end


function lineextradetails.getOrCreateLineData(lineId)
	if (datastorage.version ~= 1) then
		datastorage.version = 1

	end

	if (datastorage.lines == nil) then
		datastorage.lines = {}
	end

	if (datastorage.lines[lineId] == nil) then
		datastorage.lines[lineId] = {}
	end
	local linedetails = datastorage.lines[lineId]
	return linedetails
end

function lineextradetails.getLineData(lineId)
	if (datastorage == nil) then
		return nil
	end
	if (datastorage.lines == nil) then
		return nil
	end
	if (datastorage.lines[lineId] == nil) then
		return nil
	end
	local linedetails = datastorage.lines[lineId]
	return linedetails
end


function lineextradetails.onLineVehicleSetNextStation(eventdata)
	local trace = false
--[[
	if (eventdata.id == 70107) then
		trace = true
		_log("onLineVehicleSetNextStation found vehicle ", eventdata.id, eventdata.line )
	end
--]]

	if (eventdata.line == nil) then
		return
	end
	local lineId = eventdata.line

	local linedetails = lineextradetails.getLineData(lineId)
	if (linedetails == nil) then
		if (trace) then
			_log("onLineVehicleSetNextStation linedetails not set", eventdata.id, eventdata.line )
		end
		return
	end

	-- fast return path, if no settings are set (perf patch)
	if (linedetails.onEmptyVehicle == nil) then
		if (trace) then
			_log("onLineVehicleSetNextStation onEmptyVehicle not set", eventdata.id, eventdata.line )
		end
		return
	end
	if (trace) then
		commonapi.dmp(linedetails.onEmptyVehicle)
	end
	local tv = game.interface.getEntity(eventdata.id)
	local cargoLoad = tv.cargoLoad or {}

	if (next(cargoLoad) ~= nil) then
		if (trace) then
			_log("onLineVehicleSetNextStation still cargo on vehicle", eventdata.id, eventdata.line )
			commonapi.dmp(cargoLoad)
		end
		return
	end
	if (trace) then
		commonapi.dmp(tv)
	end

	local curStopIndex = tv.stopIndex
	local stopDataWithoutWaypoints = lineextradetails.getLineStops(lineId, false)

	if (stopDataWithoutWaypoints[curStopIndex+1] ~= nil) then
		local uniqueId = stopDataWithoutWaypoints[curStopIndex+1].uniqueId
		local onEmptyData = lineextradetails.getOnEmptyVehicleData(lineId, uniqueId)
		if (onEmptyData and onEmptyData.returnTo ~= nil) then
			if (onEmptyData.returnTo == 0) then
				_log("onLineVehicleSetNextStation found empty vehicle with onEmptyVehicle rule", eventdata.id, eventdata.line )
--[[
				commonapi_gamescript.vehiclechanges[#commonapi_gamescript.vehiclechanges+1] = {
					id = eventdata.id,
					line = eventdata.line,
					nextStation = 0
				}
--]]


				return {
					nextStation = 0
				}


--[[
				local createCmd = function()
					local cmd = api.cmd.make.setLine(eventdata.id, lineId, 0)
					api.cmd.sendCommand(cmd)
					return true
				end

				local ok, err = pcall(createCmd)
				if (ok == nil) then
					print(err)
				end
--]]
			end
		end
	else
		_log("onLineVehicleSetNextStation can't find current stop?", eventdata.id, eventdata.line )
	end
	return
end


function lineextradetails.onLineVehicleSetNextStationLater()
	for idx, data in ipairs(commonapi_gamescript.vehiclechanges) do
		_log("onLineVehicleSetNextStationLater sending setLine", data.id, data.line, data.nextStation)
		local cmd = api.cmd.make.setLine(data.id, data.line, data.nextStation)
		api.cmd.sendCommand(cmd)
	end
	commonapi_gamescript.vehiclechanges = {}
	return true
end


function lineextradetails.verifyLines()
	if (datastorage.lines == nil) then
		return
	end

	local lineIds = {}
	for lineId, lineData in pairs(datastorage.lines) do
		lineIds[#lineIds+1] = lineId
	end

	for i, lineId in ipairs(lineIds) do
		lineextradetails.verifyLineData(lineId)
	end
end

function lineextradetails.verifyLineData(lineId)
	_log("verifyLineData("..lineId..")")

	local linedetails = lineextradetails.getLineData(lineId)
	if (linedetails == nil) then
		return
	end

	if (api.engine.entityExists(lineId) == false) then
		_log("verifyLineData() removing because line doesn't exists anymore", lineId)
		datastorage.lines[lineId] = nil
		return
	end

	local listOfStops, uniqueIdList = lineextradetails.getLineStops(lineId, false)

	if (linedetails.endStations ~= nil) then
		if (listOfStops == nil) then
			linedetails.endStations = nil
		else
			local remove = {}

			for key, _ in pairs(linedetails.endStations) do
				if (uniqueIdList[key] == nil) then
					remove[#remove+1] = key
				end
			end
			for i = 1, #remove do
				local key = remove[i]
				_log("Removing non existent stop from line", key)
				linedetails.endStations[key] = nil
			end
		end
		if (linedetails.endStations ~= nil and next(linedetails.endStations) == nil) then
			_log("verifyLineData() removing endStations")
			linedetails.endStations = nil
		end
	end

	if (linedetails.onEmptyVehicle ~= nil) then
		if (listOfStops == nil) then
			linedetails.onEmptyVehicle = nil
		else
			local remove = {}
			for key, _ in pairs(linedetails.onEmptyVehicle) do
				if (uniqueIdList[key] == nil) then
					remove[#remove+1] = key
				end
			end
			for i = 1, #remove do
				local key = remove[i]
				_log("Removing non existent stop from line onEmptyVehicle", key)
				linedetails.onEmptyVehicle[key] = nil
			end
		end

		if (linedetails.onEmptyVehicle ~= nil and next(linedetails.onEmptyVehicle) == nil) then
			_log("verifyLineData() removing onEmptyVehicle")
			linedetails.onEmptyVehicle = nil
		end
	end

	if (next(datastorage.lines[lineId]) == nil) then
		_log("verifyLineData() removing lines", lineId)
		datastorage.lines[lineId] = nil
	end
end



function lineextradetails.getLineStops(lineId, waypoints)
	local lineComp = api.engine.getComponent(lineId, api.type.ComponentType.LINE)
	if (lineComp == nil) then
		return nil
	end

	local entityVisited = {}
	local getEntityIdx = function(entityId)
		if (entityVisited[entityId] == nil) then
			entityVisited[entityId] = 1
		else
			entityVisited[entityId] = entityVisited[entityId] + 1
		end
		return entityVisited[entityId]
	end

	local listOfStops = {}
	local uniqueIdList = {}

	for stopIndex, stopData in ipairs(lineComp.stops) do
		local entityId = stopData.stationGroup
		local idx = getEntityIdx(entityId)
		local uniqueId = entityId..'-'..idx
		listOfStops[#listOfStops+1] = {
			id = entityId,
			idx = idx,
			uniqueId = uniqueId,
			lineId = lineId,
		}
		uniqueIdList[uniqueId] = true

		if (waypoints == true) then
			for waypointIndex, waypointEntity in ipairs(stopData.waypoints) do
				local idxW = getEntityIdx(waypointEntity)
				local uniqueIdW = waypointEntity..'-'..idxW
				listOfStops[#listOfStops+1] = {
					id = waypointEntity,
					idx = idxW,
					uniqueId = uniqueIdW,
					lineId = lineId,
					waypoint = true,
				}
				uniqueIdList[uniqueIdW] = true
			end
		end
	end

	return listOfStops, uniqueIdList
end


function lineextradetails.getEndStations(lineId)
	if (datastorage.lines == nil) then
		return nil
	end
	if (datastorage.lines[lineId] == nil) then
		return nil
	end
	local linedetails = datastorage.lines[lineId]
	return linedetails.endStations
end




function lineextradetails.getEndStationData(lineId, uniqueId)
	local linedetails = lineextradetails.getLineData(lineId)
	if (linedetails == nil) then
		return nil
	end
	if (linedetails.endStations == nil) then
		return nil
	end
	return linedetails.endStations[uniqueId]
end


function lineextradetails.setEndStationData(lineId, uniqueId, data)
	local linedetails = lineextradetails.getOrCreateLineData(lineId)

	if (linedetails.endStations == nil) then
		linedetails.endStations = {}
	end
	linedetails.endStations[uniqueId] = data
	lineextradetails.syncData()
end


function lineextradetails.removeEndStationData(lineId, uniqueId)
	local linedetails = lineextradetails.getLineData(lineId)
	if (linedetails == nil) then
		return nil
	end
	if (linedetails.endStations ~= nil) then
		linedetails.endStations[uniqueId] = nil
	end
	lineextradetails.syncData()
end




function lineextradetails.getOnEmptyVehicleData(lineId, uniqueId)
	local linedetails = lineextradetails.getLineData(lineId)
	if (linedetails == nil) then
		return nil
	end
	if (linedetails.onEmptyVehicle == nil) then
		return nil
	end
	return linedetails.onEmptyVehicle[uniqueId]
end


function lineextradetails.setOnEmptyVehicleData(lineId, uniqueId, data)
	local linedetails = lineextradetails.getOrCreateLineData(lineId)

	if (linedetails.onEmptyVehicle == nil) then
		linedetails.onEmptyVehicle = {}
	end
	linedetails.onEmptyVehicle[uniqueId] = data
	lineextradetails.syncData()
end


function lineextradetails.removeOnEmptyVehicleData(lineId, uniqueId)
	local linedetails = lineextradetails.getLineData(lineId)
	if (linedetails == nil) then
		return nil
	end
	if (linedetails.onEmptyVehicle ~= nil) then
		linedetails.onEmptyVehicle[uniqueId] = nil
	end
	lineextradetails.syncData()
end



return lineextradetails