local InspectWindow = {}

local ui = require("commonapi2.ui2.tpfui")
local inspectlua = require("commonapi2.extlib.inspect")


local zoneutil = require "mission.zone"


InspectWindow = ui.Window:new()

InspectWindow.id =  "InspectWindow"
InspectWindow.caption = "Inspector"
InspectWindow.uilist = {}

InspectWindow.height = 500
InspectWindow.width = 500
InspectWindow._zoneid = "commonapi2_InspectWindow_zone"
InspectWindow._zoneSizes = {
	10, 20, 40, 50
}

InspectWindow._zoneSizeDefault = 2


local secondaryui = require "commonapi2.ui2.secondaryui"

local function EventTraceOutputCallback(idsource, id, name, param)
	--EventTraceOutput
	local eleid = "XCoTPF2.InspectWindow.EventTraceScrollRegion"
	local ele = api.gui.util.getById(eleid)
	if (ele ~= nil) then

		local layout = ele:getContent():getLayout()
		if (layout) then
			local itemtext = idsource.." "..id.."  "..name.. "\n"..inspectlua(param, { metatables = 0 })
			local item = api.gui.comp.TextView.new(itemtext)
			layout:addItem(item, false)

			if (layout:getNumItems() > 10) then
				local tempitem = layout:getItem(0)
				layout:removeItem(tempitem)
				tempitem:destroy()
			end
		--	layout:ensureVisible(item)
		end
	else
		secondaryui.traceGUICallback = nil
	end
end


local function EventTraceOutputStart()
	secondaryui.traceGUICallback = EventTraceOutputCallback
end

local function EventTraceOutputStop()
	secondaryui.traceGUICallback = nil
end



function InspectWindow.create()
	local o
	o = InspectWindow:new()
	o:createUIList()
	o:open()
	o:setPositionCenter()

	return o
end

function InspectWindow:refresh()
	self:createUIList()
end
function InspectWindow:onclose()
	if (game and game.interface and game.interface.setZone) then
		game.interface.setZone(self._zoneid)
	end
	--print("InspectWindow:onclose called")
end

function InspectWindow:getComboStrings(values)
	local strings = {}
	for idx, value in ipairs(values) do
		strings[idx] = tostring(value)
	end
	if (values[1] == 0) then
		strings[1] = self._("Default")
	end
	return strings
end


function InspectWindow:getComboValue(name, values, defaultIndex)
	local comboIndex = self.params[name]
	if (comboIndex == nil) then
		if (defaultIndex ~= nil) then
			comboIndex = defaultIndex
		else
			comboIndex = 0
		end
	end

	local value = values[comboIndex+1] or 0
	return value
end

function InspectWindow:createUIList()
	local uilist = {}
	local uitabs = {}
	uitabs[#uitabs+1] = {
		typ = "TabItem",
		id = "EntityTab",
		caption = "Entity",
		uilist = self:createUIListTabEntity()
	}


	uitabs[#uitabs+1] = {
		typ = "TabItem",
		id = "InspectRadiusTab",
		caption = "Entities",
		uilist = self:createUIListTabRadius()
	}


	uitabs[#uitabs+1] = {
		typ = "TabItem",
		id = "EventTab",
		caption = "Events",
		uilist = self:createUIListTabEvents()
	}


	uilist[#uilist+1] = {
		typ = "TabBar",
		id = "InspectorTabs",
		uilist = uitabs,
	}
	self.uilist = uilist
end

function InspectWindow:createUIListTabRadiusZone()
	local radius = self:getComboValue("ZoneSize", self._zoneSizes, self._zoneSizeDefault )
	local cameraPos = game.gui.getCamera()
	-- x, y, z, turn, tilt
	local pos = {
		cameraPos[1], cameraPos[2]
	}

	local num = 32
	if (radius < 20) then
		num = 8
	elseif radius < 100 then
		num = 12
	elseif radius < 160 then
		num = math.floor(radius/5)
	end

	local zonedata = {
		polygon = zoneutil.makeCircleZone(pos, radius, num),
		draw = true,
		drawColor = { .8, .8, .8, .7 },
	}
	game.interface.setZone(self._zoneid, zonedata)
	self._zonedata = {
		pos = pos,
		radius = radius
	}
end
function InspectWindow:createUIListTabRadius()
	local uilist = {}


	local uilistbuttonsZone = {}
	uilist[#uilist+1] = {
		typ = "LayoutHorizontal",
		uilist = uilistbuttonsZone
	}

	uilistbuttonsZone[#uilistbuttonsZone+1] = {
		typ = "Button",
		id = "GetZoneItems",
		caption = self._("Get Entities"),
		onclick = function(item, win)
			win:createUIListTabRadiusZone()
			win:refresh()
		end
	}

	uilistbuttonsZone[#uilistbuttonsZone+1] = {
		typ = "ComboBox",
		id = "ZoneSize",
		values = self:getComboStrings(self._zoneSizes),
		defaultIndex = self._zoneSizeDefault,
		onclick = function(item, win)
			win:createUIListTabRadiusZone()
			win:refresh()
		end
	}

	local uiTableData = {}

	local headerColumns = {
		{
			caption = "EntityId",
			width = 50,
		},
		{
			caption = _("Type"),
			width = 50,
		},
		{
			caption = "Name"
		},
	}
	local TableHeader  = {
		typ = "TableHeader",
		items = headerColumns,
		onchange = function(item, win)
			win.cursorting = item.cursorting
			win:refresh()
		end
	}
	uiTableData[#uiTableData+1] = TableHeader


	uilist[#uilist+1] = {
		typ = "Table",
		id = "EntitiesInZoneTable",
		columns = #headerColumns,
		border = false,
		uilist = uiTableData,
		size = {
			0, -100
		},
		flags = {
			Sortable = true,
			MultiSortable = true,
			ScrollFreezeTopRow = true
		},
		onclick = function(item, win)
			local index = win.params["EntitiesInZoneTable"]
			if (index >= 0 and self._zoneEntities) then
				local entityid = self._zoneEntities[index+1]
				if (entityid) then
					win.params.EntityId = tostring(entityid)
				end
			end
			win:refresh()
		end
	}

	if (self._zonedata) then
		local entities = game.interface.getEntities(self._zonedata, { includeData = false } )
		self._zoneEntities = entities
		for _, entityid in pairs( entities ) do
			self:createUIListTabRadius_CreateRow(uiTableData, entityid)
		end
	end
	return uilist
end


local ComponentTypes = api.type.ComponentType
local typeMap = {
	{ ComponentTypes.BASE_EDGE_STREET, "STREET" },
	{ ComponentTypes.BASE_EDGE_TRACK, "TRACK" },
	{ ComponentTypes.MODEL_PERSON, "PERSON" },
	{ ComponentTypes.CONSTRUCTION, "CONSTRUCTION" },
	{ ComponentTypes.ASSET_GROUP, "ASSET" },
	{ ComponentTypes.STATION, "STATION" },
	{ ComponentTypes.STATION_GROUP, "STATION_GROUP" },
	{ ComponentTypes.TOWN_BUILDING, "TOWN_BUILDING" },
	{ ComponentTypes.VEHICLE_DEPOT, "DEPOT" },
	{ ComponentTypes.SIGNAL_LIST, "SIGNAL" },
	{ ComponentTypes.SIM_BUILDING, "SIM_BUILDING" },
	{ ComponentTypes.BASE_NODE, "BASE_NODE" },
	{ ComponentTypes.TRANSPORT_VEHICLE, "TRANSPORT_VEHICLE" },
	{ ComponentTypes.RAILROAD_CROSSING, "RAILROAD_CROSSING" },
}
local function getEntityType(entityId)

	for _, typeMapEntry in ipairs(typeMap) do
		local componentData = api.engine.getComponent(entityId, typeMapEntry[1])
		if (componentData) then
			return typeMapEntry[2]
		end
	end
	return nil
end


function InspectWindow:createUIListTabRadius_CreateRow(uilist, entityId)
	local uientryId = {
		typ = "Label",
		caption = tostring(entityId),
		rownew = true,
		column = 0
	}
	uilist[#uilist+1] = uientryId

	local nameColumn = ""
	if (api.engine.entityExists(entityId)) then
		local nameComponent = api.engine.getComponent(entityId, api.type.ComponentType.NAME)
		if (nameComponent) then
			nameColumn = nameComponent.name
		end

		local type = getEntityType(entityId)
		if (type) then
			uilist[#uilist+1] = {
				typ = "Label",
				caption = type,
				column = 1
			}
		end
	end

	uilist[#uilist+1] = {
		typ = "Label",
		caption = nameColumn,
		column = 2
	}
end

function InspectWindow:createUIListTabEntity()
	local uilist = {}

	local entityId = tonumber(self.params.EntityId)


	local uilistbuttonsInput = {}
	uilist[#uilist+1] = {
		typ = "LayoutHorizontal",
		uilist = uilistbuttonsInput
	}

	uilistbuttonsInput[#uilistbuttonsInput+1] = {
		typ = "TextInput",
		id = "EntityId",
		changeonenter = true,
		onchange = function(item, win)
			win:refresh()
		end
	}


	uilistbuttonsInput[#uilistbuttonsInput+1] = {
		typ = "Button",
		caption = self._("Last selected"),
		id = "EntityGetCurrrent",
		onclick = function(item, win)
			self.params.EntityId = secondaryui.lastEntitySelected
			win:refresh()
		end
	}

	local componentTypes = getmetatable(api.type.ComponentType).__index
	local componentTypesKeys = {}

	for k in pairs(componentTypes) do
		componentTypesKeys[#componentTypesKeys+1] = k
	end
	table.sort(componentTypesKeys)

	local uilistregion = {}


	if (entityId ~= nil and entityId > 0) then

		uilist[#uilist+1] = {
			typ = "Label",
			caption = "Entity ID: "..entityId
		}



		uilist[#uilist+1] = {
			typ = "ScrollRegion",
			id = "ScrollRegion",
			border = false,
			size = {
				0, -30
			},
			uilist = uilistregion
		}
		local filterCName = self.params.EntityFilter or ""
		local filterList = {}

		local filtermap = {
			EntityFilterAccount = "ACCOUNT",
			EntityFilterLotList = "LOT_LIST",
			EntityFilterLogBook = "LOG_BOOK",
			EntityFilterModelInstanceList = "MODEL_INSTANCE_LIST"
		}

		for pkey, comtype in pairs(filtermap) do
			if (self.params[pkey] ~= false) then
				filterList[comtype] = true
			end
		end
--[[

		if (self.params.EntityFilterAccount or false) then
			filterList["ACCOUNT"] = true
		end

		if (self.params.EntityFilterLotList or false) then
			filterList["LOT_LIST"] = true
		end

		if (self.params.EntityFilterLogBook or false) then
			filterList["LOG_BOOK"] = true
		end

		if (self.params.EntityFilterModelInstanceList or false) then
			filterList["MODEL_INSTANCE_LIST"] = true
		end
--]]

		if (api.engine.entityExists(entityId)) then

			local fnEntry = function(entityId, componentName)
				if (filterList[componentName] ~= nil) then
					return
				end

				local componentTypeId = api.type.ComponentType[componentName]
				local componentData = api.engine.getComponent(entityId, componentTypeId)

				if (componentData ~= nil) then

					uilistregion[#uilistregion+1] = {
						typ = "Separator"
					}

					uilistregion[#uilistregion+1] = {
						typ = "Label",
						caption = componentName
					}

					uilistregion[#uilistregion+1] = {
						typ = "Label",
						caption = inspectlua(componentData),
						id = "data_"..componentName
					}

				end
			end

			if (filterCName ~= "" and api.type.ComponentType[filterCName] ~= nil) then
				fnEntry(entityId, filterCName)
			else
				for _, componentName in ipairs(componentTypesKeys) do
					fnEntry(entityId, componentName)
				end
			end
		end
	else
		uilist[#uilist+1] = {
			typ = "Label",
			caption = "Entity ID: <not found>"
		}
	end

	uilist[#uilist+1] = {
		typ = "Separator"
	}


	local uilistbuttons = {}
	uilist[#uilist+1] = {
		typ = "LayoutHorizontal",
		uilist = uilistbuttons
	}

	uilistbuttons[#uilistbuttons+1] = {
		typ = "Button",
		id = "Refresh",
		caption = self._("Refresh"),
		onclick = function(item, win)
			win:refresh()
		end
	}

	uilistbuttons[#uilistbuttons+1] = {
		typ = "Label",
		caption = "Filter:"
	}
	uilistbuttons[#uilistbuttons+1] = {
		typ = "TextInput",
		id = "EntityFilter",
		changeonenter = true,
		onchange = function(item, win)
			win:refresh()
		end
	}


	uilistbuttons[#uilistbuttons+1] = {
		typ = "Label",
		caption = "Hide ACCOUNT:"
	}
	uilistbuttons[#uilistbuttons+1] = {
		typ = "Checkbox",
		id = "EntityFilterAccount",
		default = true,
		onclick = function(item, win)
			win:refresh()
		end
	}
	uilistbuttons[#uilistbuttons+1] = {
		typ = "Label",
		caption = "LOT_LIST:"
	}
	uilistbuttons[#uilistbuttons+1] = {
		typ = "Checkbox",
		id = "EntityFilterLotList",
		default = true,
		onclick = function(item, win)
			win:refresh()
		end
	}
	uilistbuttons[#uilistbuttons+1] = {
		typ = "Label",
		caption = "LOG_BOOK:"
	}
	uilistbuttons[#uilistbuttons+1] = {
		typ = "Checkbox",
		id = "EntityFilterLogBook",
		default = true,
		onclick = function(item, win)
			win:refresh()
		end
	}

	uilistbuttons[#uilistbuttons+1] = {
		typ = "Label",
		caption = "MODEL_INSTANCE_LIST:"
	}

	uilistbuttons[#uilistbuttons+1] = {
		typ = "Checkbox",
		id = "EntityFilterModelInstanceList",
		default = true,
		onclick = function(item, win)
			win:refresh()
		end
	}

	--65359

	--self.uilist = uilist
	return uilist
end


function InspectWindow:createUIListTabEvents()
	local uilist = {}
	local uilistregion = {}


	uilist[#uilist+1] = {
		typ = "ScrollRegion",
		id = "EventTraceScrollRegion",
		border = true,
		size = {
			0, -30
		},
		uilist = uilistregion
	}

	uilist[#uilist+1] = {
		typ = "Separator"
	}

	local uilistbuttons = {}
	uilist[#uilist+1] = {
		typ = "LayoutHorizontal",
		uilist = uilistbuttons
	}


	uilistbuttons[#uilistbuttons+1] = {
		typ = "Button",
		id = "EventTraceStart",
		caption = self._("Start"),
		onclick = function(item, win)
			EventTraceOutputStart()
		end
	}

	uilistbuttons[#uilistbuttons+1] = {
		typ = "Button",
		id = "EventTraceStop",
		caption = self._("Stop"),
		onclick = function(item, win)
			EventTraceOutputStop()
		end
	}

	uilistbuttons[#uilistbuttons+1] = {
		typ = "Button",
		id = "ClearEventTrace",
		caption = self._("Clear"),
		onclick = function(item, win)
			local eleid = win._ugidprefix.."EventTraceScrollRegion"
			local ele = api.gui.util.getById(eleid)
			if (ele ~= nil) then
				local layout = ele:getContent():getLayout()
				local numitems = layout:getNumItems()
				if (numitems > 0) then
					for i = numitems, 1, -1 do
						local tmp = layout:getItem(i-1)
						layout:removeItem(tmp)
						tmp:destroy()
					end
				end
			end
		end
	}
	return uilist
end

return InspectWindow