
local _utils = commonapi.utils
local _ = _utils._t
local _log = commonapi.utils.logPrep("commonapi2.ui.tpfui")

--_log("Reloading TPFUI")

local tracker = require("commonapi2.ui2.tpfui_tracker")
local tpfui = {}

local UIEvent = require("commonapi2.uicommon.UIEvent")

local UIItemCreator = {

}

local UIItemDestroyer = {

}



local function _EnsureEleId_inLayoutAndRemove(layout, oldele)
	if (layout.getNumItems == nil) then
		return false;
	end
	local numitems = layout:getNumItems()
	if (numitems > 0) then
		for i = 1, numitems do
			local child = layout:getItem(i-1)
			if (child == oldele) then
				--_log("Found layout for "..oldele:getId())
				layout:removeItem(oldele)
				return true
			end
		end
	end
	return false
end

local function _EnsureEleId(ele, id)
	local oldele = api.gui.util.getById(id)
	if (oldele ~= nil) then

--		_log("_EnsureEleId old id found: ", id)
	--	commonapi.dmp(oldele)
		if (oldele.getParent ~= nil) then
			--commonapi.dmp(oldele:getParent())

			local found = false
			local parent = oldele:getParent()
			if (parent ~= nil) then
				if (found == false and parent.removeItem) then
					found = _EnsureEleId_inLayoutAndRemove(parent, oldele)
				end
				if (found == false and parent.getLayout) then
					found = _EnsureEleId_inLayoutAndRemove(parent:getLayout(), oldele)
				end
				if (found == false and parent.getContent) then
					found = _EnsureEleId_inLayoutAndRemove(parent:getContent(), oldele)
				end
			end
		end

		if (oldele.remove ~= nil) then
			--_log("_EnsureEleId remove")
			oldele:remove()
		end

		if (oldele.destroy) then
			--_log("_EnsureEleId destroy")
			oldele:destroy()
		end

	end
	ele:setId(id)
end



function UIItemDestroyer._removeLayoutItemsRecursive(win, ele, deep)
	if (ele == nil) then
		return
	end
	if (deep == nil) then
		deep = 0
	end
	--local ident = string.rep("    ", deep)
	-- print("UIItemDestroyer._removeLayoutItemsRecursive(win, ", ele:getId(), deep)

	if (ele.getNumItems ~= nil) then
		local numitems = ele:getNumItems()
		if (numitems > 0) then
			for i = numitems, 1, -1 do
				local tmp = ele:getItem(i-1)
--[[
				if (tmp.getId ~= nil) then
					print(ident.."Removing id", tmp:getId())
				end
				if (tmp.getName ~= nil) then
					print(ident.."Removing name", tmp:getName())
				end
--]]
				ele:removeItem(tmp)
				UIItemDestroyer._removeLayoutItemsRecursive(win, tmp, deep+1)

				tmp:destroy()
			end
		end
	end
	if (ele.getLayout ~= nil) then
		local layout = ele:getLayout()
		if (layout ~= nil) then
			UIItemDestroyer._removeLayoutItemsRecursive(win, layout, deep+1)
		end
	end

	if (ele.getContent ~= nil) then
		local content = ele:getContent()
		if (content ~= nil) then
			UIItemDestroyer._removeLayoutItemsRecursive(win, content, deep+1)
		end
	end

	-- if (ele.getComponent ~= nil) then
	--  	local comp = ele:getComponent()
	--  	if (comp ~= nil) then
	--  		UIItemDestroyer._removeLayoutItemsRecursive(win, comp, deep+1)
	--  	end
	-- end

end


function UIItemCreator._removeItems(win, item)
	--print("UIItemCreator._removeItems", item.id)
	if (item._oldlist ~= nil) then
		for k, child in pairs(item._oldlist) do
			if (child._uglayout ~= nil) then
				UIItemCreator._removeItems(win, child)
			end
		end
	end
	if (item._uglayout ~= nil) then
		local numitems = item._uglayout:getNumItems()
		if (numitems > 0) then
			for i = numitems, 1, -1 do
				local tmp = item._uglayout:getItem(i-1)
--[[
				if (tmp.getId ~= nil) then
					print("Removing id", tmp:getId())
				end
--]]
				item._uglayout:removeItem(tmp)
				tmp:destroy()
			end
		end
	end
end

function UIItemCreator._createitem(win, item)
	local typ = item.typ or "Label"
	local fn = UIItemCreator[typ]
	if (fn ~= nil) then
		local ele = fn(win, item)
		return ele
	end
	return nil
end

function UIItemCreator._createitems(win, parent, layout)
	if (layout == nil) then
		layout = parent._uglayout
	end
	for k, item in pairs(parent.uilist) do
		local typ = item.typ or "Label"
		local fn = UIItemCreator[typ]
		if (fn ~= nil) then
--[[
			if (item.id ~= nil) then
				print("Creating ", item.id)
			end
--]]
			local ele = fn(win, item)
			if (ele ~= false and ele ~= nil) then
				layout:addItem(ele)
			end
		end
	end
	parent._oldlist = parent.uilist
end


function UIItemCreator.ScrollRegion(self, item)
	if (item.id == nil) then
		_log("UIItemCreator.ScrollRegion missing id in ", self.id)
		return nil
	end

	local fullid = self._ugidprefix..item.id

	if (item._uglayout == nil) then
		item._uglayout = api.gui.layout.BoxLayout.new("VERTICAL")
		tracker.add(self.id, item._uglayout)

		--item._uglayout:setId(fullid.."_layout")
		--_EnsureEleId(item._uglayout, fullid.."_layout")
	end

	-- item._ugscrollcontent = api.gui.util.getById(fullid.."_scrollcontent")
	if (item._ugscrollcontent == nil) then
		item._ugscrollcontent = api.gui.comp.Component.new("commonapi2_ugscrollcontent")
		tracker.add(self.id, item._ugscrollcontent)
		--item._ugscrollcontent:setId(fullid.."_scrollcontent")
		_EnsureEleId(item._ugscrollcontent, fullid.."_scrollcontent")
		--item._ugscrollcontent:setMinimumSize(api.gui.util.Size.new(400, 400))
	end
	item._ugscrollcontent:setLayout(item._uglayout)

	local ele = api.gui.comp.ScrollArea.new(item._ugscrollcontent, "XScrollRegion")
	tracker.add(self.id, ele)
	ele:setMaximumSize(api.gui.util.Size.new(-1, -1))
	ele:setGravity(-1,-1)
	--ele:setId(fullid)
	_EnsureEleId(ele, fullid)

--	print("created "..fullid)
	UIItemCreator._createitems(self, item)
	return ele
end


function UIItemCreator.Separator(self, item)
	local ele = api.gui.comp.Component.new("HorizontalLine")
	tracker.add(self.id, ele)
	return ele
end

function UIItemCreator.Label(self, item)
	local ele = api.gui.comp.TextView.new(item.caption or "")
	tracker.add(self.id, ele)
	--ele:setText(item.caption)
	return ele
end


function UIItemCreator.Text(self, item)
	local ele = api.gui.comp.TextView.new(item.caption or "")
	if (item.id ~= nil) then
		local fullid = self._ugidprefix..item.id
		ele:setId(fullid)
	end
	tracker.add(self.id, ele)
	--ele:setText(item.caption)
	return ele
end

function UIItemCreator.Button(self, item)
	local fullid = self._ugidprefix..item.id

	local eletext
	local ele

	-- eletext = api.gui.util.getById(fullid.."_Button")
	if (eletext == nil) then
		eletext = api.gui.comp.TextView.new(item.caption or "Button")
		tracker.add(self.id, eletext)
	else
		eletext:setText(item.caption)
	end

--	ele = api.gui.util.getById(fullid)
	if (ele == nil) then
		ele = api.gui.comp.Button.new(eletext, true)
		tracker.add(self.id, ele)
		--ele:setId(fullid)
		_EnsureEleId(ele, fullid)
	end

	local this = self
	local onclickfn = function()
		this.params[item.id] = true
		UIEvent.createClick(this, item)
	end
	ele:onClick(onclickfn)
	return ele
end

function UIItemCreator.Checkbox(self, item)
	local fullid = self._ugidprefix..item.id
	local param
	if (self.params[item.id] ~= nil) then
		param = self.params[item.id]
	else
		param = item.default or false
	end

	local ele -- = api.gui.util.getById(fullid)
	if (ele == nil) then
		local cstyle = {"ui/checkbox0.tga", "ui/checkbox1.tga"}
	--	local cstyle = {"ui/design/components/checkbox_invalid@2x.tga", "ui/design/components/checkbox_valid@2x.tga"}
		if (item.checkboxstyle == "onoff") then
			cstyle = {"ui/onoff0.tga", "ui/onoff1.tga"}
		end
		ele = api.gui.comp.CheckBox.new(item.caption or "", cstyle[1], cstyle[2])
		tracker.add(self.id, ele)
		--ele:setId(fullid)
		_EnsureEleId(ele, fullid)
	else
		ele:setText(item.caption or "")
	end

	--item.caption or "")
	ele:setSelected(param, false)
	local this = self
	local onToggleFn = function(newbool)
		this.params[item.id] = newbool
		UIEvent.createClick(this, item)
	end
	ele:onToggle(onToggleFn)
	return ele
end


function UIItemCreator.TextInput(self, item)
	local fullid = self._ugidprefix..item.id
	local param = ""
	local ok

	if (self.params[item.id] ~= nil) then
		param = ""..self.params[item.id]
	elseif(item.default ~= nil) then
		param = ""..item.default or ""
	else
		param = ""
	end
	local ele
	-- = api.gui.util.getById(fullid)
	if (ele == nil) then
		ele = api.gui.comp.TextInputField.new(param)
		tracker.add(self.id, ele)
		--ele:setId(fullid)
		_EnsureEleId(ele, fullid)
		ele:setText(param or "", false)
	else
		ele:setText(param or "", false)
	end
	
	local this = self
	local onChangeFn = function(newvalue)
		if (newvalue == nil) then
			newvalue = ele:getText()
		end

		if (item.typ == "IntValue" or item.typ == "FloatValue") then
			this.params[item.id] = tonumber(newvalue)
		else
			this.params[item.id] = newvalue
		end
		UIEvent.createChange(this, item)
	end
	if (item.changeonenter ~= nil and item.changeonenter == true) then

	else
		ele:onChange(onChangeFn)
	end
	ele:onEnter(onChangeFn)

	return ele
end

UIItemCreator.IntValue = UIItemCreator.TextInput
UIItemCreator.FloatValue = UIItemCreator.TextInput


function UIItemCreator.ComboBox(self, item)
	if (item.id == nil) then
		print("UIItemRender.ComboBox missing id in ", self.id)
		return nil
	end

	local fullid = self._ugidprefix..item.id

	local param

	if (self.params[item.id] ~= nil) then
		param = self.params[item.id]
	else
		param = item.defaultIndex or 0
	end

	local ele = api.gui.comp.ComboBox.new()
	tracker.add(self.id, ele)
	--ele:setId(fullid)
	_EnsureEleId(ele, fullid)

	if type(item.values) == "table" then
		local i = 0
		for key, value in ipairs(item.values) do
			ele:addItem(value)
			i = i + 1;
		end
		if (item.values[param+1] == nil) then
			param = 0
		end
	end

	local this = self
	local onIndexChangeFn = function(newvalue)
		this.params[item.id] = newvalue
		UIEvent.createClick(this, item)
	end
	ele:onIndexChanged(onIndexChangeFn)
	ele:setSelected(param, false)
	return ele
end



function UIItemCreator.Table(self, item)
	if (item.id == nil) then
		print("UIItemCreator.Table missing id in ", self.id)
		return nil
	end
	local fullid = self._ugidprefix..item.id
	local columnCount =  item.columns or 1

	local type = "NONE"
	if (item.onchange ~= nil or item.onclick) then
		type = "SINGLE"
	end
	local ele = api.gui.comp.Table.new( columnCount, type)
	_EnsureEleId(ele, fullid)

	local row = {}

	local this = self
	local onChangeFn = function(index)
		this.params[item.id] = index
		if (item.onclick ~= nil) then
			if (index >= 0) then
				UIEvent.createClick(this, item)
				ele:select(-1, false)
			end
		else
			UIEvent.createChange(this, item)
		end
	end
	ele:onSelect(onChangeFn)

	local addRowHelper = function()
		if (#row > 0) then
			for i = 1, columnCount do
				if (row[i] == nil) then
					row[i] = api.gui.comp.TextView.new("")
				end
			end
			ele:addRow(row)
		end
		row = {}
	end

	for k, childitem in pairs(item.uilist) do
		if (childitem.typ == "TableHeader") then
			local headerElements = {}
			for k, header in ipairs(childitem.items) do
				headerElements[#headerElements+1] = api.gui.comp.TextView.new(header.caption or "")
				if (header.width) then
					ele:setMinColWidth(#headerElements-1, math.floor(header.width))
				end
			end
			ele:setHeader(headerElements)
		else
			if (childitem.rownew == true) then
				addRowHelper()
			end
			local col = (childitem.column or 0) + 1
			local childele = UIItemCreator._createitem(self, childitem)
			if (childele) then
				row[col] = childele
			end
		end
	end	-- item.uilist
	addRowHelper()
	return ele
end


function UIItemCreator.TabBar(self, item)
	if (item.id == nil) then
		print("UIItemCreator.TabBar missing id in ", self.id)
		return nil
	end

	local fullid = self._ugidprefix..item.id

	local tabelement = api.gui.util.getById(fullid)
	if (tabelement == nil) then
		tabelement = api.gui.comp.TabWidget.new("NORTH")
		tracker.add(self.id, tabelement)
		tabelement:setId(fullid)
	end
	--ele:setId(fullid)
	--_EnsureEleId(tabelement, fullid)
	item._tabelement = tabelement
	item._uglayout = tabelement:getLayout()

	for k, tabitem in pairs(item.uilist) do
		local typ = tabitem.typ or "Label"
		local caption = tabitem.caption or "TabItem"
		if (typ == "TabItem") then
			local fn = UIItemCreator[typ]
			if (fn ~= nil) then
				if (tabitem.id ~= nil) then
					--print("Creating ", tabitem.id)
				end
				local ele = fn(self, tabitem)
				if (ele ~= false and ele ~= nil) then
					tabelement:addTabText(caption, ele)
				end
			end
		end
	end

	if (self.params[item.id] ~= nil) then
		tabelement:setCurrentTab(self.params[item.id], false)
	end

	local this = self
	local onChangeFn = function(tabindex)
		this.params[item.id] = tabindex
		UIEvent.createChange(this, item)
	end
	tabelement:onCurrentChanged(onChangeFn)
--	UIItemCreator._createitems(self, item)
	return tabelement
end

function UIItemCreator.TabItem(self, item)
	if (item.id == nil) then
		print("UIItemCreator.TabItem missing id in ", self.id)
		return nil
	end

	local fullid = self._ugidprefix..item.id

	--item._uglayout = api.gui.util.getById(fullid.."_layout")
	if (item._uglayout == nil) then
		item._uglayout = api.gui.layout.BoxLayout.new("VERTICAL")
		tracker.add(self.id, item._uglayout)
		--item._uglayout:setId(fullid.."_layout")
	--	_EnsureEleId(item._uglayout, fullid.."_layout")
	end

	if (item._ugtabcontent == nil) then
		item._ugtabcontent = api.gui.comp.Component.new("commonapi2_ugtabcontent")
		tracker.add(self.id, item._ugtabcontent)
	--	item._ugtabcontent:setId(fullid.."_tabcontent")
	end
--	print("Trying set layout on", item.id)
	item._ugtabcontent:setLayout(item._uglayout)
--	print("Trying set layout done", item.id)

	UIItemCreator._createitems(self, item)
	return item._ugtabcontent
end


function UIItemCreator.LayoutHorizontal(self, item)
--[[
	if (item.id == nil) then
		print("UIItemCreator.LayoutHorizontal missing id in ", self.id)
		return nil
	end
--]]
	--local fullid = self._ugidprefix..item.id

	if (item._uglayout == nil) then
		item._uglayout = api.gui.layout.BoxLayout.new("HORIZONTAL")
		tracker.add(self.id, item._uglayout)

		--item._uglayout:setId(fullid.."_layout")
		--_EnsureEleId(item._uglayout, fullid.."_layout")
	end
	if (item._ugcontent == nil) then
		item._ugcontent = api.gui.comp.Component.new("commonapi2_layouth")
		tracker.add(self.id, item._ugcontent)
		if (item.id) then
			local fullid = self._ugidprefix..item.id
			item._ugcontent:setId(fullid)
		end
	end
	item._ugcontent:setLayout(item._uglayout)

--	print("created "..fullid)
	UIItemCreator._createitems(self, item)
	return item._ugcontent
end


local UIWindow = {
	_open = true,

	id = "",
	uilist = {},
	params = {},
	--["_"] = _G["_"] or commonapi.utils._t
	["_"] = commonapi.utils._t
}

--- UI Window creation function
-- @tparam[opt] table o A table to use while construction
-- @treturn UIWindow
function UIWindow:new(o)
	local o = o or {}

	for k, v in pairs(self) do
		if (o[k] == nil) then
			o[k] = _utils.deepcopy(v)
		end
	end

--	o._super = self
	return o
end

function UIWindow:getElementById(id)
	for k, item in pairs(self.uilist) do
		if (item.id == id) then
			return item
		end
	end
	return nil
end

function UIWindow:setPosition(x, y)
	self._ugwindow:setPosition(x, y)
end

function UIWindow:getPositionSize()
	local rect = self._ugwindow:getContentRect()
	local ret = {
		x = rect.x,
		y = rect.y,
		w = rect.w,
		h = rect.h,
	}
	if (rect.w == 0 or rect.h == 0) then
		local minsize = self._ugwindow:calcMinimumSize()
		ret.w = minsize.w
		ret.h = minsize.h
	end
	return ret
end


function UIWindow:setPositionCenter()
	if (self._ugwindow == nil) then
		return
	end
	if (self._ugwindow.close == nil) then
		-- fake window
		return
	end


	local rootele = tpfui._getRoot()
	local posSize = self:getPositionSize()

	local rootrect = rootele:getContentRect()

	local centerx = (rootrect.w / 2) - (posSize.w / 2)
	local centery = (rootrect.h / 2) - (posSize.h / 2)

	if (centerx < 0) then
		centerx = 0
	end

	if (centery < 0) then
		centery = 0
	end

	self._ugwindow:setPosition(math.floor(centerx), math.floor(centery))
end

function UIWindow:_createWindow(caption, layout)
	if (api.gui.comp.Window ~= nil) then
		if (caption == nil) then
			caption = self.id
		end
		local win = api.gui.comp.Window.new( caption, layout )
		--win:setPosition(100, 100)
		return win
	end

	local window = api.gui.comp.Component.new("Window") --CommonAPI2WindowFake")

	--window:setLayout(layout)

	local windowLayout = api.gui.layout.BoxLayout.new("VERTICAL")
	window:setLayout(windowLayout)

	local windowTitleBar = api.gui.comp.Component.new("Window::Title-bar")
	local windowTitleBarLayout = api.gui.layout.BoxLayout.new("HORIZONTAL")
	windowTitleBar:setLayout(windowTitleBarLayout)

	local windowContent = api.gui.comp.Component.new("Window::Content")
	windowLayout:addItem(windowTitleBar)
	windowLayout:addItem(windowContent)
	windowContent:setLayout(layout)



	local TitleCompLayout = api.gui.layout.BoxLayout.new("VERTICAL")
	local TitleComp = api.gui.comp.Component.new("Window::TitleComp")
	TitleComp:setLayout(TitleCompLayout)
	TitleCompLayout:addItem(api.gui.comp.TextView.new(caption))
	windowTitleBarLayout:addItem(TitleComp)
	local windowTitleBarBoxLayoutSpace = api.gui.comp.Component.new("BoxLayout::Space")

	windowTitleBarBoxLayoutSpace:setMaximumSize(api.gui.util.Size.new(-1, -1))
	windowTitleBarBoxLayoutSpace:setGravity(-1,-1)
	windowTitleBarLayout:addItem(windowTitleBarBoxLayoutSpace)



	local closebuttonIcon = api.gui.comp.ImageView.new("ui/design/window-content/close.tga")
	local closebutton = api.gui.comp.Button.new(closebuttonIcon, true)
	windowTitleBarLayout:addItem(closebutton)




	local that = self
	closebutton:onClick(function()
		that:close()
	end)

	--local windowlayout = api.gui.comp.BoxLayout()
	local menuui = api.gui.util.getById("menuUI")

	if (menuui ~= nil) then
		local windowcontainer = tpfui._findByName( menuui:getParent() , "WindowContainer", 1)
		local windowcontainerlayout = windowcontainer:getLayout()
		local rect = api.gui.util.Rect.new(30, 30, 100, 100)
		windowcontainerlayout:addItem(window, rect)
	end

	local meta = {
		__gc = function(gcself)
			if (api and api.gui and api.gui.util and gcself._ugwindow) then
				local parent = gcself._ugwindow:getParent()
				if (parent ~= nil) then
					local gclayout = parent:getLayout()
					if (gclayout ~= nil) then
						print("destroying fake window: ", gcself.id)
						if (gcself._ugwindow ~= nil) then
							gclayout:removeItem(gcself._ugwindow)
							if (gcself._uglayout ~= nil) then
								local numitems = gcself._uglayout:getNumItems()
								if (numitems > 0) then
										for i = numitems, 1, -1 do
											local item = gcself._uglayout:getItem(i-1)
											print("UIWindow __gc: Removing element", item:getId())
											gcself._uglayout:removeItem(item)

										end
									end
							end

							gcself._ugwindow:destroy()
							gcself._ugwindow = nil
							print("destroyed fake window: ", gcself.id)
						end
					end
				end
			end
		end
	}
	setmetatable(self, meta)

	return window
end

function UIWindow:_closeWindow()
	local ele = self._ugwindow
	--print("UIWindow:_closeWindow: " ..  "XCoTPF2."..self.id.."._window")

	UIEvent.createClose(self, self)

	-- as we destruct, we won't fire events, do it now:
	local ok, err = pcall(self._fireevents, self)
	if (ok == false) then
		commonapi.dmp(err)
	end

	if (self._uglayout) then
		local numitems = self._uglayout:getNumItems()
		if (numitems > 0) then
			for i = numitems, 1, -1 do
				local item = self._uglayout:getItem(i-1)
				--print("UIWindow:_closeWindow: Removing element", item:getId())
				self._uglayout:removeItem(item)
				--print("UIWindow:_closeWindow: using destroy later for element", item:getId())
				api.gui.util.destroyLater(item)
			end
		end
		--print("UIWindow:_closeWindow: Setting _uglayout id to empty string")
		self._uglayout:setId("")
	end
	self._ugwindow = nil
	self._uglayout = nil

	if (ele == nil) then
		return
	end

	--print("UIWindow:_closeWindow: Setting window id to empty string")
	ele:setId("")

	if (ele.close == nil) then
		-- Fake window
		local parent = ele:getParent()
		local layout = parent:getLayout()
		layout:removeItem(ele)

		if (api.gui.util.destroyLater ~= nil) then
			api.gui.util.destroyLater(ele)
		else
			ele:destroy()
		end
	else
--		local destroyer = function()
--			ele:remove()
--		end
--		ele:invokeLater(destroyer)


		if (api.gui.util.destroyLater ~= nil) then
--			print("UIWindow:_closeWindow: using destroyLater for window")

--			local x = require("debuggerHelper")
--			x:start()
			local parent = ele:getParent()
			local layout = parent:getLayout()
			layout:removeItem(ele)

			api.gui.util.destroyLater(ele)
		else
			ele:destroy()
		end

	end
	--print("UIWindow:_closeWindow done")

end

function UIWindow:_fixwindowsize()
	local screenSize = tpfui._getRoot():getContentRect()
	local currentContentRect = o._ugwindow:getContentRect()

	local newSize = api.gui.util.Size.new(currentContentRect.w, currentContentRect.h)

	local maxw = math.floor(screenSize.w*0.9)
	local maxh = math.floor(screenSize.h*0.9)

	if (currentContentRect.w > maxw) then
		newSize.w = maxw
	end
	if (currentContentRect.h > maxh) then
		newSize.h = maxh
	end
	if (newSize.w > -1 or newSize.h > -1) then
	--	self._ugwindow:setSize(newSize)
	end

	--local maxsize = api.gui.util.Size.new(maxw, maxh)
	--self._uglayout:setMaximumSize(maxsize)
end


function UIWindow:_renderRebuildItems()
	--local debuggerHelper = require("debuggerHelper")
	--debuggerHelper.start()

--	print(" UIWindow:_renderRebuildItems UIItemCreator._removeItems")
	UIItemCreator._removeItems(self, self)
--	print(" UIWindow:_renderRebuildItems UIItemCreator._createitems")
	UIItemCreator._createitems(self, self)
end


function UIWindow:_render()
	self._ugidprefix = "XCoTPF2."..self.id.."."
	local windowid = self._ugidprefix.."_window"

--[[
	local _ugwindow = api.gui.util.getById(windowid)

	if (_ugwindow ~= nil) then
		print("Found old window: "..windowid)
		if (_ugwindow.remove) then
			_ugwindow:remove()
		end
		if (_ugwindow.close == nil) then

		end
	end
--]]

	self._ugwindow = api.gui.util.getById(windowid)

	if (self._ugwindow) then
		if (self._ugwindow.getContent) then
			self._uglayout = self._ugwindow:getContent()
		else
			local tmplayout = self._ugwindow:getLayout()
			self._uglayout = tmplayout:getItem(1):getLayout()
		end

		local numitems = self._uglayout:getNumItems()
		if (numitems > 0) then
			for i = numitems, 1, -1 do
				local tmp = self._uglayout:getItem(i-1)
				self._uglayout:removeItem(tmp)
				tmp:destroy()
			end
		end
	end

	--self._uglayout = api.gui.util.getById(self._ugidprefix.."_layout")
	-- = api.gui.util.getById("XCoTPF2.ModSettingsWindow_eis_os_commonapi2_1._layout")

	--print(self._ugidprefix)
	if (self._ugwindow == nil) then

		if (self._uglayout == nil) then
			self._uglayout = api.gui.layout.BoxLayout.new("VERTICAL")
			--self._uglayout:setId(self._ugidprefix.."_layout")
		end
		--game.gui.window_create(self._ugidprefix.."_window", self.caption or "Test", self._ugidprefix.."_layout")
		--self._ugwindow = api.gui.comp.Window.new( self.caption, self._uglayout )
		self._ugwindow = self:_createWindow( self.caption, self._uglayout )

		local screenSize = tpfui._getRoot():getContentRect()
		local maxsize = api.gui.util.Size.new( screenSize.w ,screenSize.h )
		--commonapi.dmp(screenSize)
		--commonapi.dmp(maxsize)

		self._ugwindow:setMaximumSize(maxsize)

		self._ugwindow:setId(windowid)
		local that = self
		local closehandler = function()
			--that._ugwindow:setVisible(false, false)
			that:close()
		end

		if (self._ugwindow.onClose) then
			self._ugwindow:onClose(closehandler)
		end
		--self._ugwindow = api.gui.util.getById(self._ugidprefix.."_window")
	end
	local this = self
	local stepfn = function()
		if (this._ugwindow == nil) then
			print("stepFn called with _ugwindow nil!")
		end
		local ok, err = pcall(this._fireevents, this)
		if (ok == false) then
			print(err)
		end
		if (this._oldlist ~= nil and this._oldlist ~= this.uilist) then
			ok, err = pcall(this._renderRebuildItems, this)
			if (ok == false) then
				print(err)
			end

			--UIItemCreator._removeItems(this, this)
			--UIItemCreator._createitems(this, this)
			--local result, err = pcall(this._fixwindowsize, this)
			--if (result == false) then
			--	print(err)
			--end
		end
	end
	self._ugwindow:onStep(stepfn)

	if (self._ugwindow.setResizable) then
		self._ugwindow:setResizable(true)
	end
--	UIItemDestroyer._removeLayoutItemsRecursive(self, self._uglayout)
--	print("_removeLayoutItemsRecursive done")
	if (self._oldlist ~= self.uilist) then
--		print("UIWindow._render self._oldlist ~= self.uilist")
		local ok, err = pcall(self._renderRebuildItems, self)
		if (ok == false) then
			commonapi.dmp(err)
			print(err:getId())
		end

		--UIItemCreator._removeItems(self, self)
		--UIItemCreator._createitems(self, self)
	end

end

function UIWindow:_fireevents()
	UIEvent.fireEvents(self)

	if (self.changed == true) then
		if (self.onchange ~= nil) then
			self:onchange()
		end
		self.changed = false
	end
end

--- Open this window
--
--
--
-- Warning: Proper id of window needs to be set before use!
function UIWindow:open()
	self._open = true
	self._firstframe = true
	self:_render()
	self._ugwindow:setVisible(true, false)
end

--- Close this window
function UIWindow:close()
	self._open = false
	self:_closeWindow()
end

tpfui.Window = UIWindow



local function inspectUI(element, xprint, deep)
	local ele = element
	if (type(ele) == "string") then
		ele = api.gui.util.getById(element)
	end

	if (deep == nil) then
		deep = 0
	end

	local ident = string.rep("  ", deep)
	local ident2 = string.rep("  ", deep+1)
	if (ele == nil) then
		xprint(ident.."nil\n")
		return
	end

	local layout
	local content
	local numitems = 0
	local mt = getmetatable(ele)
	local data = {}

	if (mt ~= nil) then
		if (mt.__name ~= nil) then
			data[#data+1] = "typ = "..mt.__name
			--print("mt.__name", mt.__name)
			--if (mt.__name == "sol.UI::TextView*") then
			--	return
			--end
			if (mt.__index == nil) then
				xprint(ident..data[1] .. " BROKEN!")
				return
			end
		end
	end
	data[#data+1] = "ptr = "..tostring(ele)

	if (ele.getId ~= nil) then
		local id = ele:getId()
		if (id ~= nil) then
			if (id ~= "") then
				data[#data+1] = "id = \""..id.."\""
			end
		end
	end

	if (ele.getName ~= nil) then
		local name = ele:getName()
		if (name ~= nil) then
			if (name ~= "") then
				data[#data+1] = "name = \""..name.."\""
			end
		end
	end

	if (ele.getLayout ~= nil) then
		layout = ele:getLayout()
		if (layout == nil) then
			data[#data+1] = "layout = <nil>"
		end
	end


	if (ele.getNumItems ~= nil) then
		numitems = ele:getNumItems()
		data[#data+1] = "numItems = "..numitems
	end


	if (ele.getContent ~= nil) then
		content = ele:getContent()
		if (content == nil) then
			data[#data+1] = "layout = <nil>"
		end
	end

	if (ele.getText ~= nil) then
		local text = ele:getText()
		if (text == nil) then
			text = "nil"
		end
		data[#data+1] = "text = \""..text:sub(1, 100).."\""
	end

	xprint(ident..table.concat(data, ", "))
	if (layout ~= nil) then
		xprint(ident2.."layout = {")
		inspectUI(layout, xprint, deep+3)
		xprint(ident2.."}")
	end

	if (content ~= nil) then
		xprint(ident2.."content = {")
		inspectUI(content, xprint, deep+3)
		xprint(ident2.."}")
	end

	if (numitems > 0) then
		xprint(ident2.."items = {")
		if (ele.getItem ~= nil) then
			for i = 1, numitems do
				local child = ele:getItem(i-1)
				inspectUI(child, xprint, deep+3)
			end
		else
			xprint(ident2.." <missing getItem function>")
		end
		xprint(ident2.."}")
	end
	--xprint(ident.."\n")
end

--[[
package.loaded["commonapi2.uicommon.interface"] = nil
interface = require("commonapi2.uicommon.interface")
interface.inspect(interface._getRoot())
require("commonapi2.uicommon.interface").inspect("inGameConsole")

x = api.gui.util.getById("inGameConsole")
x:getLayout():getItem(0)
commonapi.dmp( x:getLayout():getItem(0):getLayout():getItem(1) )

win = api.gui.util.getById("inGameConsole")
win:getParent():getParent():getParent():getLayout():getItem(0)
--]]

tpfui.inspect = function(ele, filename)
	local xprint = print
	local file, errstring
	if (type(filename) == "string") then
		file, errstring = io.open(filename, "a")
		if (file ~= nil) then
			xprint = function(var)
				file:write(var.."\n")
			end
		else
			print(errstring)
		end
	end
	inspectUI(ele, xprint)
	if (file ~= nil) then
		file:close()
	end
end

local function rounddec(f, digits)
	local muliply = 10^(digits or 0)
	return math.floor(f * muliply + 0.5) / muliply
end


tpfui.getUIScale = function()
	local ele
	local ret = 1.0
	ele = api.gui.util.getById("menu.finances")

	if (ele) then
		local crect = ele:getContentRect()
		ret = rounddec( crect.h / 85, 2)
	end
	return ret
end


tpfui.getUICenter = function()
	local rootele = tpfui._getRoot()
	local scale = tpfui.getUIScale()
	if (rootele) then
		local rootrect = rootele:getContentRect()
		local centerx = math.floor(rootrect.w / 2 * scale)
		local centery = math.floor(rootrect.h / 2 * scale)
		return centerx, centery
	end
	return nil
end

tpfui._getRoot = function()
	local ele -- = api.gui.util.getById("inGameConsole")

	--if (ele == nil) then
		ele = api.gui.util.getById("menuUI")
	--end

	if (ele == nil) then
		ele = api.gui.util.getById("mainView")
	end

	if (ele == nil) then
		return nil
	end

	local parent = ele:getParent()
	while(parent) do
		ele = parent
		parent = ele:getParent()
	end
	return ele
end

-- commonapi.ui._findByName( api.gui.util.getById("menuUI"):getParent(), "WindowContainer", 1)
tpfui._findByName = function(parent, name, maxdeep)
	maxdeep = maxdeep or 1
	if (parent.getLayout == nil) then
		return nil
	end
	local layout = parent:getLayout()

	if (layout == nil) then
		return nil
	end

	local numitems = layout:getNumItems()
	for i = 1, numitems do
		local child = layout:getItem(i-1)
		if (child ~= nil) then
			if (child.getName ~= nil) then
				if (child:getName() == name) then
					return child
				end
			end
			if (maxdeep > 1) then
				local ret = tpfui._findByName(child, name, maxdeep-1)
				if (ret ~= nil) then
					return ret
				end
			end
		end
	end
	return nil
end

tpfui._getParentWindowOLD = function(ele)
	if (type(ele) == "string") then
		ele = api.gui.util.getById(ele)
	end
	if (ele == nil) then
		return
	end
	local parent = ele:getParent()
	while(parent) do
		local mt = getmetatable(ele)
		print(">>"..mt.__name.."<<")
		if (mt.__name == "sol.UI::Window*") then
			break
		end
		ele = parent
		parent = ele:getParent()
	end
	return ele
end


tpfui._getParentWindow = function(ele)
	if (type(ele) == "string") then
		ele = api.gui.util.getById(ele)
	end
	if (ele == nil) then
		return
	end
	local curele = ele
	while(curele) do
		local mt = getmetatable(curele)
		--print(">>"..mt.__name.."<<")
		if (mt.__name == "sol.UI::Window*") then
			return curele
		end
		if (curele.getName ~= nil) then
			if (curele:getName() == "Window") then
				return curele
			end
		end
		if (curele.getParent == nil) then
			break
		end
		curele = curele:getParent()
	end
	return nil
end


tpfui._getParentWindowX = function(ele)
	if (type(ele) == "string") then
		ele = api.gui.util.getById(ele)
	end
	if (ele == nil) then
		return
	end
	local parent = ele:getParent()
	while(parent) do
		local mt = getmetatable(ele)
		--print(">>"..mt.__name.."<<")
		if (mt.__name == "sol.UI::Window*") then
			break
		end
		if (ele.getName ~= nil and ele.setId ~= nil) then
			if (ele:getName() == "Window") then
				local oldEleId = ele:getId()
				if (oldEleId == "") then
					ele:setId("__temporary_id")
					print(">>Trying id hack to get window<<")
					local newele = api.gui.util.getById("__temporary_id")
					ele:setId("")
					local newmt = getmetatable(newele)
					if (newmt.__name == "sol.UI::Window*") then
						return newele
					end
				end
			end
		end
		ele = parent
		parent = ele:getParent()
	end
	return ele
end


---
-- Creates a entry in the commonapi mod menu
-- @tparam string id unique id for this entry
-- @tparam string caption
-- @tparam function fn callback function
function tpfui.createModMenuEntry(id, caption, fn)
end

---
-- Removes id from commonapi mod menu
function tpfui.removeModMenuEntry(id)
end

--- Message Box Creator
-- @tparam string id unique id for this msgbox. (overwrites previous message with same id)
-- @tparam string text
-- @tparam string caption
function tpfui.msgbox(id, text, caption, buttons)

end

function tpfui._activate()

end
return tpfui
