---- UI Paramater creating
-- @author Oskar Eisemuth
-- @copyright 2017-2021
-- @module commonapi.uiparameter
local uiparameter = {
	version = 2  -- Version
}

local _log = commonapi.utils.logPrep("commonapi2.uiparameter")
--print("Loading commonapi.uiparameter")
local filter = commonapi.filter

local settingsloaded = false
local liststyle = 0
local compressui = false

-- controlChar is 2 bytes long
local controlChar = "©"

local function loadMySettings()
	if (settingsloaded) then
		return
	end
	
	liststyle = commonapi.settings.get("eis_os_commonapi2_1", "liststyle") or 0
	compressui = commonapi.settings.get("eis_os_commonapi2_1", "compressui")
	if (compressui == nil) then
		compressui = false
	end
	_log("liststyle", liststyle)
	_log("compressui", compressui)
	settingsloaded = true
end

local function _entrySorter(a, b)
	if (a.data == nil or b.data == nil) then
		return false
	end
	local AyearFrom = (a.data.yearFrom or 0)
	local ByearFrom = (b.data.yearFrom or 0)
	if ( AyearFrom < ByearFrom ) then
--		print("compare <:", a.filename.."="..AyearFrom, b.filename.."="..ByearFrom)
		return true
	elseif ( AyearFrom == ByearFrom ) then
		if ( a.id < b.id ) then
--			print("byid  :", a.filename.."="..AyearFrom, b.filename.."="..ByearFrom)
			return true
		end
--	else
--		print("compare  :", a.filename.."="..AyearFrom, b.filename.."="..ByearFrom)
	end
	return false
end

local function dmpentries(entries)
	for i, entry in ipairs(entries) do
		print(i, entry.filename)
	end
end

----
--- UI Parameter interface
-- @section uiinterface

local defaultNames = {
	track = {
		["standard.lua"] = _("Standard"),
		["high_speed.lua"] = _("High-speed"),
	}, 
	bridge = {
		["rail_wood_old.lua"] = _("Wooden bridge"),
		["rail_stone_old.lua"] = _("Stone"),
		["rail_stone_new.lua"] = _("Concrete bridge"),
		["rail_iron_old.lua"] = _("Steel"),
		["rail_iron_new.lua"] = _("Steel bridge"),

		["road_wood_old.lua"] =  _("Wooden bridge"),
		["road_stone_old.lua"] = _("Stone"),
		["road_stone_new.lua"] = _("Concrete bridge"),
		["road_iron_old.lua"] =  _("Steel"),
		["road_iron_new.lua"] =  _("Steel bridge"),
	}
}

--[[
--------------------------------
-- UI Objects                 --
--------------------------------
--]]

--- Base ui parameter Object
-- @type UITypeSelector
local UITypeSelector = {
	filters = { filter.visibleInUI },
	_filters = { filter.visibleInUI }
}

--- Constructor
-- @param[opt] o object
function UITypeSelector:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o.filters = table.copy(o.filters)
	o._filters = table.copy(o._filters)
	o._super = self
	return o
end


--- Adds a new filter to the filter list
-- @tparam function filter
function UITypeSelector:addFilter(filter)
	self.filters[#self.filters+1] = filter
end


function UITypeSelector:_getFailBackEntries()
	--if (api.gui)
end



--- Returns a filtered list of items in the repository
-- @see commonapi.repos
-- @treturn {commonapi.repos.repoentry,...} Entries
function UITypeSelector:getEntries()
	-- For TPF2, only support internal filters
	local entries = commonapi.repos[self.typ].getEntriesFiltered(self._filters)

	if (liststyle == 3) then
		table.sort(entries, _entrySorter)
	end
	return entries
end

function UITypeSelector:getEntryCaption(entry)
	local name, cnt
	local data = entry.data or {}
	if (data.shortUIName ~= nil) then
		name = data.shortUIName
	elseif (entry.filename ~= nil) then
		local shortUINames = defaultNames[self.typ] or {}
		if (shortUINames[entry.filename] ~= nil) then
			name = shortUINames[entry.filename]
		end
	end
	
	if (name == nil and data.name ~= nil) then
		name = data.name
	end
	if (name == nil) then
		name = "?"
	end
	name = tostring(name)
	--print(type(_(name)), _(name) % {}, "="..name)
	
	if (data.speedLimit ~= nil) then
		local tname = commonapi.utils.translateString(entry.modid, name)
		if (tname) then
			tname, cnt = string.gsub(tname, "(%d+%s*km/h%s*)", "")
		end
		name = tname .. " (" .. tostring(math.floor(data.speedLimit*3.6)) .. ")"
	end
	return name
end



--- Creates ui params
-- @tparam table params Parameter table reference
function UITypeSelector:createUIParams(params)
	loadMySettings()
--[[	if (liststyle == 1) then
		return self:_createUIParamsAllEntries(params)
	end
	return self:_createUIParamsByYears(params)
--]]
	return self:_createUIParamsAllEntries(params)
end

local function compactEntries(entryNames)
	local compareNames = {}
	local shorterName = {}
	for i = 1, #entryNames do
		compareNames[i] = string.gsub(entryNames[i], "%(.+%)$",
			function(n, other)
				shorterName[i] = n
				return ""
			end
		)
	end
	for i = 1, #entryNames do
		local prev = compareNames[i-1] or ""
		if (prev ~= "") then
			if (prev == compareNames[i]) then
				entryNames[i] = shorterName[i]
			end
		end
	end
end


function UITypeSelector:_createAllEntryNames()
	local entries = self:getEntries()
	local entryNames = {}

	for i, entry in ipairs(entries) do
		entryNames[#entryNames+1] = self:getEntryCaption(entry)
	end

	if (compressui == true) then
		compactEntries(entryNames)
	end

	if (#entryNames == 0) then
		entryNames[1] = "<NO DATA FOUND>"
	end

	if (self._postRunHack ~= true) then
		if (self.typ == "track") then
			entryNames[1] = controlChar.."T "..entryNames[1]
		end
	end
	return entryNames
end

function UITypeSelector:_createUIParamsAllEntries(params)
	local entryNames = self:_createAllEntryNames()

	params[#params+1] = {
		key = self.key,
		name = self.caption,
		values = entryNames,
		uiType = "COMBOBOX",
	}
end

function UITypeSelector:_createUIParamsByYears(params)
	local entries = self:getEntries()
	local entryNames = {}
	local entryYearFrom = {}
	local entryYearTo = {}


	local years = {}
	local yearIdx = {}
	for i, entry in ipairs(entries) do
		local yearFrom = (entry.data.yearFrom or 0)
		local yearTo = (entry.data.yearTo or 0)
		years[yearFrom] = yearFrom
		years[yearTo] = yearTo
		
		local idx = #entryNames+1
		entryNames[idx] = self:getEntryCaption(entry)
		entryYearFrom[idx] = yearFrom
		entryYearTo[idx] = yearTo
	end
	
	for i, year in pairs(years) do
		yearIdx[#yearIdx+1] = year
	end
	table.sort(yearIdx)
	
	local function _filteredNamesByName(yearFrom, yearTo)
		local entryNamesFiltered = {}
		for i, entryName in ipairs(entryNames) do
--			if (entryYearFrom[i] <= yearFrom and (entryYearTo[i] == 0 or entryYearTo[i] >= yearTo) ) then
				entryNamesFiltered[i] = entryName
--			else
--				entryNamesFiltered[i] = "" -- hidden
--			end
		end
		if (compressui == true) then
			compactEntries(entryNamesFiltered)
		end
		
		if (#entryNamesFiltered == 0) then
			
		end
		return entryNamesFiltered
	end
	
	
	for i = 1, #yearIdx do
		local yearFrom = yearIdx[i] or 0
		local yearTo = yearIdx[i+1] or 0
		params[#params+1] = {
			key = self.key,
			name = self.caption,
			values = _filteredNamesByName(yearFrom, yearTo),
			yearFrom = yearFrom,
			yearTo = yearTo,
			uiType = "COMBOBOX",
		}
	end
end



--- Get result from user selection
-- @tparam table params Parameter table of updateFn
-- @treturn commonapi.repos.repoentry A repository entry
function UITypeSelector:getSelection(params)
	local entries = self:getEntries()
	local entry
	local idx = (params[self.key] or 0) +1
	entry = entries[idx]
	if (entry == nil) then
		entry = entries[1]
	end
	return table.copy(entry)
end


function UITypeSelector:_getEntriesAsList()
	local entries = self:getEntries()
	local list = {}
	for i, entry in ipairs(entries) do
		list[i] = entry.filename
	end
	return list
end


----
-- Default filter list
-- @table filters


---- TrackSelector Class
-- @type TrackSelector
local TrackSelector = UITypeSelector:new({
	typ = "track",
	key = "trackType",
	caption = _("Track type"),
})
TrackSelector.filters = { filter.visibleInUI, filter._settingsTracks }

---
-- Param key
-- @string[opt="trackType"] key Key of catenary

---
-- Caption
-- @string[opt="Translated Track type"] caption Caption Text


---- Track and Catenary Selector
-- @type TrackCatenarySelector

local TrackCatenarySelector = TrackSelector:new({
	catenary_key = "catenary",
	catenary_caption = _("Catenary"),
})

--- Creates ui params including catenary
function TrackCatenarySelector:createUIParams(params)
	TrackSelector.createUIParams(self, params)
	params[#params+1] = {
		key = self.catenary_key,
		name = self.catenary_caption,
		values = { _("No"), _("Yes") },
		yearFrom = (game.config.trackCatenaryYearFrom or 1910),
		defaultIndex = 1,
	}
end

--- Get result from user selection (catenary).
function TrackCatenarySelector:getCatenarySelection(params)
	local catenary = (params[self.catenary_key] or 0) == 1
	return catenary
end

--- Get result from user selection (track)
-- @function getSelection
-- @tparam table params Parameter table of updateFn
-- @treturn commonapi.repos.repoentry A repository entry

---
-- Param key
-- @string[opt="trackType"] key Key of param

---
-- Caption
-- @string[opt="Translated Track type"] caption Caption Text

---
-- Key to be used for catenary
-- @string[opt="catenary"] catenary_key Key of catenary

---
-- Caption to be used for catenary
-- @string[opt="Translated Catenary"] catenary_caption Text of catenary

----
--- StreetSelector
-- @type StreetSelector
local StreetSelector = UITypeSelector:new({
	typ = "street",
	key = "streetType",
	caption = commonapi.utils._t("Street"),
})
StreetSelector.filters = { filter.visibleInUI, filter._settingsStreet }

---
-- Param key
-- @string[opt="streetType"] key Key of param

---
-- Caption
-- @string[opt="Translated Street"] caption Caption Text

----
--- BridgeSelector
-- @type BridgeSelector

local BridgeSelector = UITypeSelector:new({
	typ = "bridge",
	key = "bridgeType",
	caption = commonapi.utils._t("Bridge"),
})

---
-- Param key
-- @string[opt="bridgeType"] key Key of catenary

---
-- Caption
-- @string[opt="Translated Bridge"] caption Caption Text

----
--- RailBridgeSelector
-- Rail bridge selector
-- @type RailBridgeSelector
local RailBridgeSelector = BridgeSelector:new()

---
-- Param key
-- @string[opt="bridgeType"] key Key of catenary

---
-- Caption
-- @string[opt="Translated Bridge"] caption Caption Text

----
-- Default filter list, only selects bridges that carries rail
-- @table filters
RailBridgeSelector.filters = { filter.visibleInUI, filter.bridgeCarryRail }




----
--- RoadBridgeSelector
-- Road bridge selector
-- @type RoadBridgeSelector
local RoadBridgeSelector = BridgeSelector:new()

---
-- Param key
-- @string[opt="bridgeType"] key Key of catenary

---
-- Caption
-- @string[opt="Translated Bridge"] caption Caption Text

---
-- Default filters for only selects bridges that carries road.
RoadBridgeSelector.filters = { filter.visibleInUI, filter.bridgeCarryRoad }



----
-- @section end

local function _commonModifyParam(params, classObj, o)
	o = o or {}
	local obj = classObj:new(o)
	local newparams = {}
	obj:createUIParams(newparams)

	if (o.selectionlist ~= nil) then
		local selectionlist = o.selectionlist
		local list = obj:_getEntriesAsList()


		if (#list < 1) then
			_log("ERROR UI Parameter creation failed because _getEntriesAsList is empty, should not happen if native dll is loaded!")
			return obj
		end
--[[
		local m = math.max(#selectionlist, #list)
		for i = 1, m do
			selectionlist[i] = list[i]
		end
--]]
		-- Always use dynamic created list
		for i = 1, #selectionlist do
			selectionlist[i] = nil
		end
		selectionlist[1] = list[1]
		local mt = {
			__len = function(tbl)
				local entries = obj:getEntries()
				return #entries
			end,
			__index = function(tbl, index)
				_log("Using dynamic created entries")
				local entries = obj:getEntries()
				local entry = entries[index] or entries[1]
				return entry.filename
			end,
			__ipairs = function(tbl, i)
				local entries = obj:getEntries()
				local function ipairs_iter(t, k)
					if (k == nil) then
						k = 0
					end
					k = k + 1
					local v = entries[k]
					if (v ~= nil) then
						return k, v.filename
					end
				end
				return ipairs_iter, tbl, nil
			end
		}
		mt.__pairs = mt.__ipairs
		setmetatable(selectionlist, mt)
	end
-- moved here, so empty list won't crash game
	uiparameter.replace(params, newparams)
	return obj
end

----
--- UI Parameter interface
-- @section uiinterface

--- Create a track and catenary selector uiparamater
-- @param[opt] o object
-- @treturn TrackCatenarySelector
function uiparameter.createTrackCatenary(o)
	return TrackCatenarySelector:new(o)
end

--- Create a track selector uiparamater
-- @param[opt] o object
-- @treturn TrackSelector
function uiparameter.createTrack(o)
	return TrackSelector:new(o)
end

--- Create a street selector uiparamater
-- @param[opt] o object
-- @treturn StreetSelector
function uiparameter.createStreet(o)
	return StreetSelector:new(o)
end


--- Create a rail bridge selector uiparamater
-- @param[opt] o object
-- @treturn RailBridgeSelector
function uiparameter.createRailBridge(o)
	return RailBridgeSelector:new(o)
end

--- Create a road bridge selector uiparamater
-- @param[opt] o object
-- @treturn RoadBridgeSelector
function uiparameter.createRoadBridge(o)
	return RoadBridgeSelector:new(o)
end


--- Modify a track and catenary ui paramater
-- @param params reference to parameters
-- @param[opt] o object (use key "selectionlist" for overwriting a target list of filenames )
-- @treturn TrackCatenarySelector
function uiparameter.modifyTrackCatenary(params, o)
	return _commonModifyParam(params, TrackCatenarySelector, o)
end

--- Modify a rail bridge ui paramater
-- @param params reference to parameters
-- @param[opt] o object (use key "selectionlist" for overwriting a target list of filenames )
-- @treturn RailBridgeSelector
function uiparameter.modifyRailBridge(params, o)
	return _commonModifyParam(params, RailBridgeSelector, o)
end


--- Replaces ui parameters with new parameters
-- @tparam table params Parameter table reference (modified)
-- @tparam table newparams Parameter table of replacements
-- @treturn table params reference
function uiparameter.replace(params, newparams)
	local result = {}
	local bykeys = {}
	
	for i, p in ipairs (newparams) do
		local t = bykeys[p.key]
		if (t == nil) then
			t = {}
			bykeys[p.key] = t
		end
		t[#t+1] = p
	end
	
	for i, p in ipairs (params) do
		local key = p.key
		if (bykeys[key] ~= nil) then
			if (bykeys[key] == 1) then
				
			else
				for n, np in ipairs (bykeys[key]) do
					result[#result+1] = np
				end
				bykeys[key] = 1
			end
		else
			result[#result+1] = p
		end
	end
	
	for key, data in pairs(bykeys) do
		if (data == 1) then
		
		else
			for n, np in ipairs (data) do
				result[#result+1] = np
			end
		end
	end
	
	local m = math.max(#params, #result)
	for i = 1, m do
		params[i] = result[i]
	end
	return params
end



function uiparameter._prepPostRun()
	_log("_prepPostRun: Patching construction")
	local allConstructions = api.res.constructionRep.getAll()

	local uiparameterTracks = uiparameter.createTrack({ _postRunHack = true})
	local trackEntries = uiparameterTracks:_createAllEntryNames()

	local conTypes = api.type.enum.ConstructionType
	for __, conIdName in pairs(allConstructions) do
		local con = api.res.constructionRep.get(api.res.constructionRep.find(conIdName))
		if (con.type == conTypes.TOWN_BUILDING or con.type == conTypes.NONE) then
			--_log("skipping ", conIdName)
		else
			for pidx, p in ipairs(con.params) do
				if (#p.values > 0) then
					local firstvalue = p.values[1]
					if (firstvalue:sub(1,2) == controlChar) then
						if (firstvalue:sub(3,4) == "T ") then
							p.values = table.copy(trackEntries)
							p.values[1] = p.values[1].." "..controlChar
						end
					end
				end
			end
		end
	end

	_log("_prepPostRun: Patching construction done")
end


if (commonapi._native ~= nil) then
	return uiparameter
else
	-- Not supported without native dll
	--return false
	return uiparameter
end

